import os
import pandas as pd
import warnings
from typing import Optional
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types, convert_to_float, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_supply_chain_map_areas(areas: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of areas based on the given area information.

    This function takes a DataFrame of areas with their country, region and center latitude and longitude, along with an API key, 
    and creates a map of areas using the Log-hub Supply Chain Map service. 

    Parameters:
    areas (pd.DataFrame): A pandas DataFrame containing areas information.
        Each row should contain:
        - id (number): Identifier.
        - searchId (str): Area search id.
        - country (str):  Country name or country code of the area.
        - region (str): Region name.
        - name (str): Area name.
        - layer (str): Layer name for the area.
        - quantity (number): Numerical value that can be used to determine the color of an area on the map.
        - continent (str): Country and area continent.
        - countryRegionOne, countryRegionTwo (str): Continent's subregion and second subregion where country is located.
        - countryName (str): Country name.
        - centerLat, centerLong (number): Area's center point latitude and longitude.
        - population (number): Area population.
        - areaKm2 (number): Area in km2

    parameters (dict): A dictionary containg parameter 'detailLevel' (boolean). 

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 'workspaceId' (str) and 'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.
    Returns:
    pd.DataFrame: A pandas DataFrame containg the passed area information and its length [km]. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    mandatory_columns = {'country': 'str', 'region': 'str'}
    optional_columns = {
        'searchId': 'str', 'name': 'str', 'layer': 'str', 'continent': 'str', 'countryRegionOne': 'str', 'countryRegionTwo': 'str', 'countryName': 'str'
        }
    optional_floats = ['id', 'quantity', 'centerLat', 'centerLong', 'population', 'areaKm2']

    # Validate and convert data types
    areas = validate_and_convert_data_types(areas, mandatory_columns, 'mandatory', 'areas')
    if not areas is None:
        areas = validate_and_convert_data_types(areas, optional_columns, 'optional', 'areas')
        if not areas is None:
            areas = convert_to_float(areas, optional_floats, 'optional')
            areas = convert_df_to_dict_excluding_nan(areas, optional_floats)
    if areas is None:
        return None

    url = create_url("supplychainmapareas")
    
    headers = create_headers(api_key)
    payload = {
        "areaData": areas,
        "parameters": parameters
    }
    
    payload = save_scenario_check(save_scenario, payload, "supplychainmapareas")
    
    response_data = post_method(url, payload, headers, "supply chain map areas")
    if response_data is None:
        return None
    else:
        areas_df = pd.DataFrame(response_data['areaResult'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return areas_df

def forward_supply_chain_map_areas_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapAreas.xlsx')
    areas_df = pd.read_excel(data_path, sheet_name='areas', usecols='A:O')

    parameters={
        'detailLevel': False
    }
    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'mergeWithExistingScenario': False,
        'scenarioName': 'Your scenario name'
    }

    return {'areas': areas_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import logging
from typing import Optional, Tuple
import warnings
logging.basicConfig(level=logging.INFO)
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import convert_dates, validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_freight_shipment_emissions_road(addresses: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform forward freight emissions by road on a list of addresses.

    This function takes a DataFrame of addresses and an API key, and performs forward
    freight emissions by road using the Log-hub service. 

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing the addresses for calculating a CO2 footprint based on your shipments or route legs defined as separate shipments.
        Required columns and their types:
        - shipmentId (str): Unique identifier for each shipment.
        - shipmentDate (str): Dates related to the shipment (Format: YYYY-MM-DD)
        - fromCountry (str): Country code of the address from which a shipment should be sent.
        - fromState (str): State code of the address from which a shipment should be sent.
        - fromPostalCode (str): Postal code of the address from which a shipment should be sent.
        - fromCity (str): City name of the address from which a shipment should be sent.
        - fromStreet (str): Street name with house number of the address from which a shipment should be sent.
        - toCountry (str): Country code of the address to which a shipment should arrive at.
        - toState (str): State code of the address to which a shipment should arrive at.
        - toPostalCode (str): Postal code of the address to which a shipment should arrive at.
        - toCity (str): City name of the address to which a shipment should arrive at.
        - toStreet (str): Street name with house number of the address to which a shipment should arrive at.
        - isRefrigirated (str): A YES/NO option that specifies whether the content being transferred through the shipments is refrigirated or not. If not specified will be taken as NO.
        - weight (number): The weight of the shipment.

    parameters (dict): A dictionary containing parameters:
        - vehicleType: "van(0-3.5)", "truck", "truckUrbanTruck", "truckMGV", "truckHGV", "truckRigid(3.5-7.5)", "truckRigid(7.5-12)", "truckRigid(12-20)" , "truckRigid(20-26)", "truckRigid(26-32)", "truckArticulated(3.5-34)", "truckArticulated(34-40)", "truckArticulated(40-44)", "truckArticulated(44-60)", "truckArticulated(60-72)", "truckGeneral", "truckAutoCarrier", "truckDray", "truckExpedited", "truckFlatbed", "truckHeavybulk", "truckLTL", "truckMixed", "truckMoving", "truckPackage", "truckSpecialized", "truckTanker" or "truckTL" 
        - fuelType: enum "diesel", "petrol", "hybrid", "CNG", "LPG", "pluginHybrid", "electricity" or "other"
        - weightUnit: enum "kilograms" or "lbs"
        - emissionStandard: enum "EURO_5" or :EURO_6"

    api_key (str): The Log-hub API key for accessing the freight emissions service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Optional[Tuple[pd.DataFrame, pd.DataFrame]]: One pandas DataFrame contains the original address information along 
                  with the calculated CO2 emissions, and the second one contains shipments with no CO2 emission calculated. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
        
    addresses_mandatory_columns = {'shipmentId': 'str', 'shipmentDate': 'str', 'fromCountry': 'str', 'toCountry': 'str', 'weight': 'float'}
    addresses_optional_columns = {'fromState': 'str', 'fromPostalCode': 'str', 'fromCity': 'str', 'fromStreet': 'str', 'toState': 'str', 'toPostalCode': 'str', 'toCity': 'str', 'toStreet': 'str', 'isRefrigirated': 'str'}

    # Validate and convert data types
    addresses = validate_and_convert_data_types(addresses, addresses_mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
        addresses = validate_and_convert_data_types(addresses, addresses_optional_columns, 'optional', 'addresses')
    if not addresses is None:
        addresses = convert_dates(addresses, ['shipmentDate'])
    if addresses is None:
        return None
    
    url = create_url("co2emissionsroad")
    
    headers = create_headers(api_key)

    payload = {
        'freightShipmentEmissionsByRoad': addresses.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "co2 emissions road")
    if response_data is None:
        return None
    else:
        freight_emissions_df = pd.DataFrame(response_data['freightShipmentEmissionOutputStandardRoad'])
        not_evaluated_df  = pd.DataFrame(response_data['notEvaluatedShipmentsStandardRoad'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return freight_emissions_df, not_evaluated_df

def forward_freight_shipment_emissions_road_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'CO2RoadAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:O').fillna("")
    parameters = {
        "vehicleType": "truck",
        "fuelType": "diesel",
        "weightUnit": "kilograms",
        "emissionStandard": "EURO_6"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_freight_shipment_emissions_road(coordinates: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse freight emissions by road on a list of coordinates.

    This function takes a DataFrame of coordinates and an API key, and performs forward
    freight emissions by road using the Log-hub service. 

    Parameters:
    coordinates (pd.DataFrame): A pandas DataFrame containing the coordinates for calculating a CO2 footprint based on your shipments or route legs defined as separate shipments.
        Required columns and their types:
        - shipmentId (str): Unique identifier for each shipment.
        - shipmentDate (str): Dates related to the shipment (Format: YYYY-MM-DD)
        - fromLatitude (number): The latitude of the coordinate from which a shipment should be sent.
        - fromLongitude(number): The longitude of the coordinate from which a shipment should be sent.
        - toLatitude (number): The latitude of the coordinate to which a shipment should arrive at.
        - toLongitude (numebr): The longitude of the coordinate to which a shipment should arrive at.
        - isRefrigirated (str): A YES/NO option that specifies whether the content being transferred through the shipments is refrigirated or not. If not specified will be taken as NO.
        - weight (number): The weight of the shipment.

    parameters (dict): A dictionary containing parameters:
        - vehicleType: "van(0-3.5)", "truck", "truckUrbanTruck", "truckMGV", "truckHGV", "truckRigid(3.5-7.5)", "truckRigid(7.5-12)", "truckRigid(12-20)" , "truckRigid(20-26)", "truckRigid(26-32)", "truckArticulated(3.5-34)", "truckArticulated(34-40)", "truckArticulated(40-44)", "truckArticulated(44-60)", "truckArticulated(60-72)", "truckGeneral", "truckAutoCarrier", "truckDray", "truckExpedited", "truckFlatbed", "truckHeavybulk", "truckLTL", "truckMixed", "truckMoving", "truckPackage", "truckSpecialized", "truckTanker" or "truckTL" 
        - fuelType: enum "diesel", "petrol", "hybrid", "CNG", "LPG", "pluginHybrid", "electricity" or "other"
        - weightUnit: enum "kilograms" or "lbs"
        - emissionStandard: enum "EURO_5" or :EURO_6"

    api_key (str): The Log-hub API key for accessing the reverse freight emissions service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Optional[Tuple[pd.DataFrame, pd.DataFrame]]: One pandas DataFrame contains the original coordinates information along 
                  with the calculated CO2 emissions, and the second one contains shipments with no CO2 emission calculated. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    coordinates_mandatory_columns = {'shipmentId': 'str', 'shipmentDate': 'str', 'fromLatitude': 'float', 'fromLongitude': 'float', 'toLatitude': 'float', 'toLongitude': 'float', 'weight': 'float'}
    coordinates_optional_columns = {'isRefrigirated': 'str'}

    # Validate and convert data types
    coordinates = validate_and_convert_data_types(coordinates, coordinates_mandatory_columns, 'mandatory', 'coordinates')
    if not coordinates is None:
        coordinates = validate_and_convert_data_types(coordinates, coordinates_optional_columns, 'optional', 'coordinates')
    if not coordinates is None:
        coordinates = convert_dates(coordinates, ['shipmentDate'])
    if coordinates is None:
        return None
    url = create_url("reverseco2emissionsroad")
    
    headers = create_headers(api_key)

    payload = {
        'freightShipmentEmissionsByRoad': coordinates.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "reverse co2 emissions road")
    if response_data is None:
        return None
    else:
        freight_emissions_df = pd.DataFrame(response_data['freightShipmentEmissionOutputReverseRoad'])
        not_evaluated_df  = pd.DataFrame(response_data['notEvaluatedShipmentsReverseRoad'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return freight_emissions_df, not_evaluated_df

def reverse_freight_shipment_emissions_road_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'CO2RoadReverse.xlsx')
    coordinates_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:I').fillna("")
    parameters = {
        "vehicleType": "truck",
        "fuelType": "diesel",
        "weightUnit": "kilograms",
        "emissionStandard": "EURO_6"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'coordinates': coordinates_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import requests
import logging
import time
from save_to_platform import get_entities_link

def create_url(app_name):
    """
    Creates an url for the given application.
    """
    DEFAULT_LOG_HUB_API_SERVER = "https://production.supply-chain-apps.log-hub.com"
    LOG_HUB_API_SERVER = os.getenv('LOG_HUB_API_SERVER', DEFAULT_LOG_HUB_API_SERVER)
    url = f"{LOG_HUB_API_SERVER}/api/applications/v1/"+app_name

    return url

def create_headers(api_key):
    
    headers = {
        "accept": "application/json",
        "authorization": f"apikey {api_key}",
        "content-type": "application/json"
    }

    return headers
def post_method(url, payload, headers, app_name):

    max_retries = 3
    retry_delay = 15  # seconds

    for attempt in range(max_retries):
        try:
            response = requests.post(url, json=payload, headers=headers)
            if response.status_code == 200:
                response_data = response.json()
                return response_data
            elif response.status_code == 429:
                logging.info(f"Rate limit exceeded. Retrying in {retry_delay} seconds.")
                time.sleep(retry_delay)
            else:
                logging.error(f"Error in {app_name} API: {response.status_code} - {response.text}")
                return None
        except requests.exceptions.RequestException as e:
            logging.error(f"Request failed: {e}")
            if attempt < max_retries - 1:
                logging.info(f"Retrying in {retry_delay} seconds.")
                time.sleep(retry_delay)

    logging.error("Max retries exceeded.")
    return None

def get_method(api_server, url, headers, app_name):

    api_server = "https://production.supply-chain-apps.log-hub.com"
    
    retry_delay = 15  # seconds
    timeout = 3600
    start_time = time.time()
    while True:
        response = requests.get(api_server + url, headers = headers)
        if response.status_code == 200:
            response_data = response.json()
            if response_data['calculationRunning']:
                elapsed_time = time.time() - start_time
                if elapsed_time > timeout:
                    logging.error("Timeout of one hour is reached with the calculation still running. Try running the calculation again.")
                    return None
                
                logging.info(f"Calculation is still running with the progress {response_data['progress']}%.")
                time.sleep(retry_delay)
            else:
                return response_data
        else:
            logging.error(f"Error in {app_name} API: {response.status_code} - {response.text}")
            return None
    
def get_workspace_entities(save_scenario, api_key):

    DEFAULT_LOG_HUB_API_SERVER = "https://production.supply-chain-apps.log-hub.com"
    LOG_HUB_API_SERVER = os.getenv('LOG_HUB_API_SERVER', DEFAULT_LOG_HUB_API_SERVER)
    url = f"{LOG_HUB_API_SERVER}/api/v1/workspace/"+ save_scenario['workspaceId'] + "/entitiesWithTables"

    max_retries = 3
    retry_delay = 15  # seconds

    for attempt in range(max_retries):
        try:
            response = requests.get(url, headers={"authorization": f"apikey {api_key}"})
            if response.status_code == 200:
                response_data = response.json()
                links = get_entities_link(response_data['data'], save_scenario['scenarioName'])
                return links
            elif response.status_code == 429:
                logging.info(f"Rate limit exceeded. Retrying in {retry_delay} seconds.")
                time.sleep(retry_delay)
            else:
                logging.error(f"Error in Entities With Tables API: {response.status_code} - {response.text}")
                return None
        except requests.exceptions.RequestException as e:
            logging.error(f"Request failed: {e}")
            if attempt < max_retries - 1:
                logging.info(f"Retrying in {retry_delay} seconds.")
                time.sleep(retry_delay)

    logging.error("Max retries exceeded.")
    return None




import os
import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_milkrun_optimization(depots: pd.DataFrame, vehicle_types: pd.DataFrame, pickup_and_delivery: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform milkrun optimization based on depots, vehicle types, and pickup and delivery orders.

    This function takes three DataFrames representing depots, vehicle types, and pickup and delivery orders, 
    along with a parameters dictionary, an API key and parameters for saving the scenario. It performs milkrun optimization using the Log-hub service.

    Parameters:
    depots (pd.DataFrame): DataFrame containing depot information.
        Columns:
        - name (str): Name of the depot.
        - country (str): Country code.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.

    vehicle_types (pd.DataFrame): DataFrame containing vehicle type information.
        Columns:
        - type (str): Vehicle Type.
        - availableVehicles (number): The number of available vehicles per vehicle type.
        - startDepot (str): Start depot name of the vehicle type.
        - endDepot (str): End depot name of the vehicle type.
        - averageSpeed (number): Average speed.
        - maxRouteStops (number): The maximum number of stops per vehicle.
        - maxRouteLength (number): The maximum length per route. 
        - maxRouteDuration (number): The maximum route duration.
        - maxCapacityWeight (number): The maximum weight that can be loaded by a vehicle of the corresponding vehicle type. 
        - maxCapacityVolume (number): The maximum volume that can be loaded by a vehicle of the corresponding vehicle type. 
        - maxCapacityPallets (number):  The maximum number of pallets that can be loaded by a vehicle of the corresponding vehicle type.
        - fixedCosts (number): The fixed costs per vehicle in case of usage. 
        - costsPerStop (number): The costs per vehicle stop.
        - costsPerDistanceUnit (number): The costs per distance unit. 

     pickup_and_delivery (pd.DataFrame): DataFrame containing pickup and delivery information.
        Columns:
        - name (str): The name of the order. 
        - country (str): Country code.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - weight (number): The weight of the order. 
        - volume (number): The volume of the order.
        - pallets (number): The number of pallets of the order.
        - pickupDelivery (str): Pickup/Delivery.
        - depot (str):  The depot name the order has to be delivered to.
        - vehicleType (str): All valid vehicle types for the order separated by semicolon.
        - serviceTime (number): Stop duration per order in minutes.
        - startTimeWindow (str): Start of the delivery/pickup time window per order.
        - endTimeWindow (str): End of the delivery/pickup time window per order.
        - externalCosts (number): External costs for order. 

    parameters (Dict): Dictionary containing parameters durationUnit and distanceUnit.

    api_key (str): Log-hub API key for accessing the milkrun optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]: Five pandas DataFrames containing route overview, route details, external orders, input map routes with addresses information and input map routes with coordinates information. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    # Define expected columns and data types for each DataFrame
    depot_mandatory_columns = {
        'country': 'str', 'name': 'str'
    }
    depot_optional_columns = {
        'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'
    }

    vehicle_types_mandatory_columns = {
        'type': 'str', 'availableVehicles': 'float',  'startDepot': 'str', 'endDepot': 'str', 'averageSpeed': 'float', 'maxRouteStops': 'float', 'maxRouteLength': 'float', 'maxRouteDuration': 'float', 'maxCapacityWeight': 'float', 'maxCapacityVolume': 'float', 'maxCapacityPallets': 'float', 'fixedCosts': 'float', 'costsPerStop': 'float', 'costsPerDistanceUnit': 'float'
    }   
    pickup_and_delivery_mandatory_columns = {
        'name': 'str', 'country': 'str', 'weight': 'float', 'volume': 'float', 'pallets': 'float', 'depot': 'str', 'serviceTime': 'float', 'startTimeWindow': 'str', 'endTimeWindow': 'str', 'externalCosts': 'float'
    }
    pickup_and_delivery_optional_columns = {
        'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'vehicleType': 'str', 'pickupDelivery': 'str'
    }

    # Perform validation and conversion for each DataFrame
    depots = validate_and_convert_data_types(depots, depot_mandatory_columns, 'mandatory', 'depots')
    if not depots is None:
        depots = validate_and_convert_data_types(depots, depot_optional_columns, 'optional', 'depots')
    vehicle_types = validate_and_convert_data_types(vehicle_types, vehicle_types_mandatory_columns, 'mandatory', 'vehicle types')
    pickup_and_delivery = validate_and_convert_data_types(pickup_and_delivery, pickup_and_delivery_mandatory_columns, 'mandatory', 'pickup and delivery')
    if not pickup_and_delivery is None:
        pickup_and_delivery = validate_and_convert_data_types(pickup_and_delivery, pickup_and_delivery_optional_columns, 'optional', 'pickup and delivery')

    # Exit if any DataFrame validation failed
    if any(df is None for df in [depots, vehicle_types, pickup_and_delivery]):
        return None

    url = create_url("milkrunoptimization")
    
    headers = create_headers(api_key)

    payload = {
        "depots": depots.to_dict(orient='records'),
        "vehicleTypes": vehicle_types.to_dict(orient='records'),
        "customers": pickup_and_delivery.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "milkrun optimization")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['droppedCustomers'])
        input_map_routes_df = pd.DataFrame(response_data['inputMapRoutes'])
        input_map_routes_geocodes_df = pd.DataFrame(response_data['inputMapRoutesGeocodes'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df, input_map_routes_df, input_map_routes_geocodes_df

def forward_milkrun_optimization_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MilkrunSampleDataAddresses.xlsx')
    depots_df = pd.read_excel(data_path, sheet_name='depots', usecols='A:G', dtype={'postalCode': str})
    vehicle_types_df = pd.read_excel(data_path, sheet_name='vehicleTypes', usecols='A:O')
    pickup_and_delivery_df= pd.read_excel(data_path, sheet_name='pickupAndDelivery', usecols='A:Q', dtype={'postalCode': str})

    parameters = {
        "distanceUnit": "km",
        "durationUnit": "min"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'depots': depots_df, 'vehicleTypes': vehicle_types_df, 'pickupAndDelivery': pickup_and_delivery_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_milkrun_optimization(depots: pd.DataFrame, vehicle_types: pd.DataFrame, pickup_and_delivery: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform milkrun optimization based on depots, vehicle types, and pickup and delivery orders.

    This function takes three DataFrames representing depots, vehicle types, and pickup and delivery orders, 
    along with a parameters dictionary, an API key and parameters for saving the scenario. It performs milkrun optimization using the Log-hub service.

    Parameters:
    depots (pd.DataFrame): DataFrame containing depot information.
        Columns:
        - name (str): Name of the depot.
        - latitude (number): Latitude of the depot.
        - longitude (number): Longitude of the depot.

    vehicle_types (pd.DataFrame): DataFrame containing vehicle type information.
        Columns:
        - type (str): Vehicle Type.
        - availableVehicles (number): The number of available vehicles per vehicle type.
        - startDepot (str): Start depot name of the vehicle type.
        - endDepot (str): End depot name of the vehicle type.
        - averageSpeed (number): Average speed.
        - maxRouteStops (number): The maximum number of stops per vehicle.
        - maxRouteLength (number): The maximum length per route. 
        - maxRouteDuration (number): The maximum route duration.
        - maxCapacityWeight (number): The maximum weight that can be loaded by a vehicle of the corresponding vehicle type. 
        - maxCapacityVolume (number): The maximum volume that can be loaded by a vehicle of the corresponding vehicle type. 
        - maxCapacityPallets (number):  The maximum number of pallets that can be loaded by a vehicle of the corresponding vehicle type.
        - fixedCosts (number): The fixed costs per vehicle in case of usage. 
        - costsPerStop (number): The costs per vehicle stop.
        - costsPerDistanceUnit (number): The costs per distance unit. 

     pickup_and_delivery (pd.DataFrame): DataFrame containing pickup and delivery information.
        Columns:
        - name (str): The name of the order. 
        - latitude (number): Latitude of the customer.
        - longitude (number): Longitude of the customer.
        - weight (number): The weight of the order. 
        - volume (number): The volume of the order.
        - pallets (number): The number of pallets of the order.
        - pickupDelivery (str): Pickup/Delivery.
        - depot (str):  The depot name the order has to be delivered to.
        - vehicleType (str): All valid vehicle types for the order separated by semicolon.
        - serviceTime (number): Stop duration per order in minutes.
        - startTimeWindow (str): Start of the delivery/pickup time window per order.
        - endTimeWindow (str): End of the delivery/pickup time window per order.
        - externalCosts (number): External costs for order. 

    parameters (Dict): Dictionary containing parameters durationUnit and distanceUnit.

    api_key (str): Log-hub API key for accessing the reverse milkrun optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]: Four pandas DataFrames containing route overview, route details, external orders and input map routes with coordinates information. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    # Define expected columns and data types for each DataFrame
    depot_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float'
    }

    vehicle_types_mandatory_columns = {
        'type': 'str', 'availableVehicles': 'float',  'startDepot': 'str', 'endDepot': 'str', 'averageSpeed': 'float', 'maxRouteStops': 'float', 'maxRouteLength': 'float', 'maxRouteDuration': 'float', 'maxCapacityWeight': 'float', 'maxCapacityVolume': 'float', 'maxCapacityPallets': 'float', 'fixedCosts': 'float', 'costsPerStop': 'float', 'costsPerDistanceUnit': 'float'
    }   
    pickup_and_delivery_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float', 'weight': 'float', 'volume': 'float', 'pallets': 'float', 'depot': 'str', 'serviceTime': 'float', 'startTimeWindow': 'str', 'endTimeWindow': 'str', 'externalCosts': 'float'
    }
    pickup_and_delivery_optional_columns = {
        'vehicleType': 'str', 'pickupDelivery': 'str'
    }

    # Perform validation and conversion for each DataFrame
    depots = validate_and_convert_data_types(depots, depot_mandatory_columns, 'mandatory', 'depots')
    vehicle_types = validate_and_convert_data_types(vehicle_types, vehicle_types_mandatory_columns, 'mandatory', 'vehicle types')
    pickup_and_delivery = validate_and_convert_data_types(pickup_and_delivery, pickup_and_delivery_mandatory_columns, 'mandatory', 'pickup and delivery')
    if not pickup_and_delivery is None:
        pickup_and_delivery = validate_and_convert_data_types(pickup_and_delivery, pickup_and_delivery_optional_columns, 'optional', 'pickup and delivery')

    # Exit if any DataFrame validation failed
    if any(df is None for df in [depots, vehicle_types, pickup_and_delivery]):
        return None

    url = create_url("reversemilkrunoptimization")
    
    headers = create_headers(api_key)

    payload = {
        "depots": depots.to_dict(orient='records'),
        "vehicleTypes": vehicle_types.to_dict(orient='records'),
        "customers": pickup_and_delivery.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "reverse milkrun optimization")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['droppedCustomers'])
        input_map_routes_geocodes_df = pd.DataFrame(response_data['inputMapRoutesGeocodes'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df, input_map_routes_geocodes_df

def reverse_milkrun_optimization_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MilkrunSampleDataReverse.xlsx')
    depots_df = pd.read_excel(data_path, sheet_name='depots', usecols='A:D')
    vehicle_types_df = pd.read_excel(data_path, sheet_name='vehicleTypes', usecols='A:O')
    pickup_and_delivery_df= pd.read_excel(data_path, sheet_name='pickupAndDelivery', usecols='A:N')

    parameters = {
        "distanceUnit": "km",
        "durationUnit": "min"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'depots': depots_df, 'vehicleTypes': vehicle_types_df, 'pickupAndDelivery': pickup_and_delivery_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_advanced_center_of_gravity(customers: pd.DataFrame, sources: pd.DataFrame, fixed_centers: pd.DataFrame, product_groups: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Calculate fixed center of gravity based on a list of customers and predefined fixed centers, with an extension of setting product groups and their sources. 

    This function takes a DataFrame of customers with weights and product group, a DataFrame of fixed centers (if any), a DataFrame of sources of the products, a DataFrame of product groups, and a set of parameters, along with an API key, and performs fixed center of gravity calculation using the Log-hub service.

    Parameters:
    customers (pd.DataFrame): A pandas DataFrame containing customers with their weights and product group information.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name. Must have at least 1 character.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - productGroup (str): Product group required from the customer.
        - weight (float): Weight.
        - assignedCenter (str): Center assigned to a customer.
        - assignedSource (str): Source of the product group.

    sources (pd.DataFrame): A pandas DataFrame containing information about sources of the product groups.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name. Must have at least 1 character.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - availableProductGroups (str): Product groups available in the named source.
        This DataFrame can be empty.

    fixed_centers (pd.DataFrame): A pandas DataFrame containing predefined fixed centers.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name. Must have at least 1 character.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        This DataFrame can be empty.
    
    product_groups (pd.DataFrame): A pandas DataFrame with the product groups data.
        Each row should contain:
        - id (number): Identifier.
        - productGroupName (str): Name of the product group.
        - serviceLevel (number): Percenatege of the total weight that must be within the given distance.
        - serviceLevelDistance (number): The distance in which the given service level must be located.

    parameters (Dict): A dictionary containing parameters: distanceUnit (enum: "km" or "mi"), minmaximumCenters (minimum number of centers) and maxmaximumCenters (maximum number of centers) integer numbers that must be equal or smaller than the amount of customers, and inboundOutboundFactor (integer number).

    api_key (str): The Log-hub API key for accessing the advanced center of gravity service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Returns three pandas DataFrames. First one contains assigned centers and their coordinates, second one describes the inbound flows from sources to centers and the third one shows flows from centers to customers. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    
    customer_mandatory_columns = {'country': 'str',  'weight': 'float', 'productGroup': 'str'}
    customer_optional_columns = {'name': 'str', 'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'assignedCenter': 'str', 'assignedSource': 'str'}
    sources_optional_columns = {'id': 'float', 'name': 'str', 'country': 'str', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    fixed_center_optional_columns = {'id': 'float', 'name': 'str', 'country': 'str', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    product_groups_mandatory_columns = {'productGroupName': 'str', 'serviceLevel': 'float', 'serviceLevelDistance': 'float'}

    # Validate and convert data types for customers and fixed centers
    customers = validate_and_convert_data_types(customers, customer_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customer_optional_columns, 'optional', 'customers')
    sources = validate_and_convert_data_types(sources, sources_optional_columns, 'optional', 'sources')
    fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_optional_columns, 'optional', 'fixed centers') if not fixed_centers.empty else fixed_centers
    product_groups = validate_and_convert_data_types(product_groups, product_groups_mandatory_columns, 'mandatory', 'product groups')
    if any(df is None for df in [customers, sources, fixed_centers, product_groups]):
        return None
    
    url = create_url("advancedcenterofgravity")
    
    headers = create_headers(api_key)

    payload = {
        "customers": customers.to_dict(orient='records'),
        "sources": sources.to_dict(orient='records'),
        "fixedCenters": fixed_centers.to_dict(orient='records') if not fixed_centers.empty else [],
        "productGroups": product_groups.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "advanced center of gravity")
    if response_data is None:
        return None
    else:
        assigned_centers_df = pd.DataFrame(response_data['assignedCenters'])
        inbound_df = pd.DataFrame(response_data['inbound'])
        outbound_df = pd.DataFrame(response_data['outbound'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_centers_df, inbound_df, outbound_df
            
def forward_advanced_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'AdvancedCoGAddresses.xlsx')
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:K', dtype={'postalCode': str})
    sources_df = pd.read_excel(data_path, sheet_name='sources', usecols='A:H', dtype={'postalCode': str})
    fixedCenters_df = pd.read_excel(data_path, sheet_name='fixedCenters', usecols='A:G', dtype={'postalCode': str})
    product_groups_df = pd.read_excel(data_path, sheet_name='productGroups', usecols='A:D')

    parameters = {
        "distanceUnit": "km",
        'minmaximumCenters': 1,
        'maxmaximumCenters': 50,
        'inboundOutboundFactor': 8
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'customers': customers_df, 'sources': sources_df, 'fixedCenters': fixedCenters_df, 'productGroups': product_groups_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_advanced_center_of_gravity(customers: pd.DataFrame, sources: pd.DataFrame, fixed_centers: pd.DataFrame, product_groups: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Calculate reverse fixed center of gravity based on a list of customers and their weights, and predefined fixed centers.

    This function takes a DataFrame of customers with weights, a DataFrame of fixed centers, and a set of parameters,
    along with an API key, and performs reverse fixed center of gravity calculation using the Log-hub service.

    Parameters:
    customers (pd.DataFrame): A pandas DataFrame containing customers with their weights.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the customer location.
        - longitude (float): Longitude of the customer location.
        - productGroup (str): Product group required from the customer.
        - weight (float): Weight.
        - assignedCenter (str): Center assigned to a customer.
        - assignedSource (str): Source of the product group.

    sources (pd.DataFrame): A pandas DataFrame containing products' sources information.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the fixed center location.
        - longitude (float): Longitude of the fixed center location.
        - availableProductGroups (str): Product groups available in the named source.
        This DataFrame can be empty.

    fixed_centers (pd.DataFrame): A pandas DataFrame containing predefined fixed centers.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the fixed center location.
        - longitude (float): Longitude of the fixed center location.

    product_groups (pd.DataFrame): A pandas DataFrame with the product groups data.
        Each row should contain:
        - id (number): Identifier.
        - productGroupName (str): Name of the product group.
        - serviceLevel (number): Percenatege of the total weight that must be within the given distance.
        - serviceLevelDistance (number): The distance in which the given service level must be located.

    parameters (Dict): A dictionary containing parameters: distanceUnit (enum: "km" or "mi"), minmaximumCenters (minimum number of centers) and maxmaximumCenters (maximum number of centers) integer numbers that must be equal or smaller than the amount of customers, and inboundOutboundFactor (integer number).

    api_key (str): The Log-hub API key for accessing the reverse advanced center of gravity service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Returns three pandas DataFrames. First one contains assigned centers and their coordinates, second one describes the inbound flows from sources to centers and the third one shows flows from centers to customers. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    customer_mandatory_columns = {
        'latitude': 'float', 'longitude': 'float', 'weight': 'float', 'productGroup': 'str'
    }
    customer_optional_columns = {'name': 'str', 'id': 'float', 'assignedCenter': 'str', 'assignedSource': 'str'}
    sources_optional_columns = {'name': 'str', 'id': 'float', 'latitude': 'float', 'longitude': 'float', 'availableProductGroups': 'str'}
    fixed_center_optional_columns = {'name': 'str', 'id': 'float', 'latitude': 'float', 'longitude': 'float'}
    product_groups_mandatory_columns = product_groups_mandatory_columns = {'productGroupName': 'str', 'serviceLevel': 'float', 'serviceLevelDistance': 'float'}

    # Validate and convert data types for customers and fixed centers
    customers = validate_and_convert_data_types(customers, customer_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customer_optional_columns, 'optional', 'customers')
    sources = validate_and_convert_data_types(sources, sources_optional_columns, 'optional', 'sources')
    fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_optional_columns, 'optional', 'fixed centers') if not fixed_centers.empty else fixed_centers
    product_groups = validate_and_convert_data_types(product_groups, product_groups_mandatory_columns, 'mandatory', 'product groups')
    if any(df is None for df in [customers, sources, fixed_centers, product_groups]):
        return None

    url = create_url("reverseadvancedcenterofgravity")
    
    headers = create_headers(api_key)

    payload = {
        "customers": customers.to_dict(orient='records'),
        "sources": sources.to_dict(orient='records'),
        "fixedCenters": fixed_centers.to_dict(orient='records') if not fixed_centers.empty else [],
        "productGroups": product_groups.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "reverse advanced center of gravity")
    if response_data is None:
        return None
    else:
        assigned_centers_df = pd.DataFrame(response_data['assignedCenters'])
        inbound_df = pd.DataFrame(response_data['inbound'])
        outbound_df = pd.DataFrame(response_data['outbound'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_centers_df, inbound_df, outbound_df

def reverse_advanced_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'AdvancedCoGReverse.xlsx')
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:H')
    sources_df = pd.read_excel(data_path, sheet_name='sources', usecols='A:E')
    fixedCenters_df = pd.read_excel(data_path, sheet_name='fixedCenters', usecols='A:D')
    product_groups_df = pd.read_excel(data_path, sheet_name='productGroups', usecols='A:D')

    parameters = {
        "distanceUnit": "km",
        'minmaximumCenters': 1,
        'maxmaximumCenters': 50,
        'inboundOutboundFactor': 8
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'customers': customers_df, 'sources': sources_df, 'fixedCenters': fixedCenters_df, 'productGroups': product_groups_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

if __name__ == "__main__":

    sample_data = forward_advanced_center_of_gravity_sample_data()
    customers_df = sample_data['customers']
    sources_df = sample_data['sources']
    fixed_centers_df = sample_data['fixedCenters']
    product_groups_df = sample_data['productGroups']
    parameters =sample_data['parameters']
    api_key = "14d872084c972dd5087bee9b6fa1593a5061a2e7"
    save_scenario = sample_data['saveScenarioParameters']
    save_scenario['saveScenario'] = True
    save_scenario['workspaceId'] = "dde088e344d7f19a669623c862cb144c4f9370dc"
    save_scenario['scenarioName'] = "Forward advanced cog"

    assigned_centers_df, inbound_df, outbound_df = forward_advanced_center_of_gravity(customers_df, sources_df, fixed_centers_df, product_groups_df, parameters, api_key, save_scenario, show_buttons=True)


import os
import pandas as pd
import logging
import warnings
from typing import Optional
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types, convert_to_float, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_supply_chain_map_sea_routes(addresses: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of sea routes based on the given route UN/LOCODES.

    This function takes a DataFrame of routes UN/LOCODES with their quantity and route restrictions, along with an API key, 
    and creates a map of sea routes using the Log-hub Supply Chain Map service. 

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing sea route UN/LOCODES with their quantity and route restrictions.
        Each row should contain:
        - id (number): Identifier.
        - fromName (str): Name for the start location of the sea route.
        - fronUnLocode (str): UN/LOCODE for the start location of the sea route.
        - toName (str): Name for the end location of the sea route.
        - toUnLocode (str): UN/LOCODE for the end location of the sea route.
        - quantity (number): Numerical value of goods being transferred through a specific sea route.
        - routingOptions (str): Restricting moving through channels, straits and passages based on the passed restrictions. 
                                Currently supported are Suez channel (as suez), Panama channel (as panama), Malacca strait (as malacca),
                                Gibraltar strait (as gibraltar), Dover strait (as dover), Bering strait (as bering), 
                                Magellan strait (as magellan), Bab-el-Mandeb strait (as babelmandeb), Kiel channel (as kiel), 
                                Corinth channel (as corinth), Northwest passage (as northwest) and Northeast passage (as northeast).
                                If multiple restrictions are named, comma or semicolon can be used as separators.

    parameters (dict): A dictionary containing parameter 'distanceUnit' (enum 'km' or 'mi' or 'nm'). 

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 
                        'workspaceId' (str) and 'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.
    Returns:
    pd.DataFrame: A pandas DataFrame containg the sea route UN/LOCODES with their quantity and route restriction options, 
                  along with the parsed latitude and longitude, parsed country name and distance.

    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    mandatory_columns = {'fromName': 'str', 'fromUnLocode': 'str', 'toName': 'str', 'toUnLocode': 'str'}
    optional_columns = {'routingOptions': 'str'}
    optional_floats = ['id', 'quantity']
    # Validate and convert data types
    addresses = validate_and_convert_data_types(addresses, mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
         addresses = validate_and_convert_data_types(addresses, optional_columns, 'optional', 'addresses')
         if not addresses is None:
              addresses = convert_to_float(addresses, optional_floats, 'optional')
              addresses = convert_df_to_dict_excluding_nan(addresses, optional_floats)
    if addresses is None:
        return None

    url = create_url("supplychainmapsearoutes")
    
    headers = create_headers(api_key)
    payload = {
        "seaRoutesAddressesData": addresses,
        "parameters": parameters
    }
    
    payload = save_scenario_check(save_scenario, payload, "supplychainmapsearoutes")
    response_data = post_method(url, payload, headers, "supply chain map sea routes")
    if response_data is None:
        return response_data
    else:
        route_result_df = pd.DataFrame(response_data['seaRoutes'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_result_df
            

def forward_supply_chain_map_sea_routes_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapSeaRoutesAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:G')

    parameters = {
        'distanceUnit': 'nm'
    }
    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_supply_chain_map_sea_routes(coordinates: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of sea routes based on the given route latitude and longitude.

    This function takes a DataFrame of routes coordinates pairs with their quantity and route restrictions, along with an API key, 
    and creates a map of sea routes using the Log-hub Supply Chain Map service. 

    Parameters:
    coordinates (pd.DataFrame): A pandas DataFrame containing sea route coordinates pairs with their quantity and route restrictions.
        Each row should contain:
        - id (number): Identifier.
        - fromName (str): Name for the start location of the sea route.
        - fromLatitude (number): Latitude of the start location.
        - fromLongitude (number): Longitude of the start location.
        - toName (str): Name for the end location of the sea route.
        - toLatitude (str): Latitude of the end location.
        - toLongitude (number): Longitude of the end location.
        - quantity (number): Numerical value of goods being transferred through a specific sea route.
        - routingOptions (str): Restricting moving through channels, straits and passages based on the passed restrictions. 
                                Currently supported are Suez channel (as suez), Panama channel (as panama), Malacca strait (as malacca),
                                Gibraltar strait (as gibraltar), Dover strait (as dover), Bering strait (as bering), 
                                Magellan strait (as magellan), Bab-el-Mandeb strait (as babelmandeb), Kiel channel (as kiel), 
                                Corinth channel (as corinth), Northwest passage (as northwest) and Northeast passage (as northeast).
                                If multiple restrictions are named, comma or semicolon can be used as separators.

    parameters (dict): A dictionary containing parameter 'distanceUnit' (enum 'km' or 'mi' or 'nm'). 

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 
                        'workspaceId' (str) and 'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containg the sea route coordinates pairs with their quantity and route restriction options, 
                  along with the parsed country name, parsed UN/LOCODES and distance.

    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])
        
    mandatory_columns = {'fromName': 'str', 'fromLatitude':'float', 'fromLongitude': 'float', 'toName': 'str', 'toLatitude': 'float', 'toLongitude': 'float'}
    optional_columns = {'routingOptions': 'str'}
    optional_floats = ['id', 'quantity']
    # Validate and convert data types
    coordinates = validate_and_convert_data_types(coordinates, mandatory_columns, 'mandatory', 'coordinates')
    if not coordinates is None:
         coordinates = validate_and_convert_data_types(coordinates, optional_columns, 'optional', 'coordinates')
         if not coordinates is None:
              coordinates = convert_to_float(coordinates, optional_floats, 'optional')
              coordinates = convert_df_to_dict_excluding_nan(coordinates, optional_floats)
    if coordinates is None:
        return None

    url = create_url("reversesupplychainmapsearoutes")
    
    headers = create_headers(api_key)
    payload = {
        "seaRoutesCoordinatesData": coordinates,
        "parameters": parameters
    }
    
    payload = save_scenario_check(save_scenario, payload, "reversesupplychainmapsearoutes")
    response_data = post_method(url, payload, headers, "reverse supply chain map sea routes")
    if response_data is None:
                return None
    else:
        route_result_df = pd.DataFrame(response_data['seaRoutes'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_result_df
            

def reverse_supply_chain_map_sea_routes_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapSeaRoutesReverse.xlsx')
    coordinates_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:I')

    parameters = {
        'distanceUnit': 'nm'
    }
    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'coordinates': coordinates_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_center_of_gravity(addresses: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Calculate center of gravity based on a list of addresses and their weights.

    This function takes a DataFrame of addresses with weights and a set of parameters,
    along with an API key, and performs center of gravity calculation using the Log-hub service.

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing addresses with their weights.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - weight (float): Weight.

    parameters (Dict): A dictionary containing parameters like numberOfCenters (number) 
                       and distanceUnit (enum: "km" or "mi").
    
    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    api_key (str): The Log-hub API key for accessing the center of gravity service.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: A tuple of two pandas DataFrames. The first DataFrame 
                                       contains the assigned addresses with their respective centers, 
                                       and the second DataFrame contains the details of the centers.
                                       Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
        
    mandatory_columns = {'country': 'str', 'weight': 'float'}
    optional_columns = {'id': 'float', 'name': 'str', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    
    # Validate and convert data types
    addresses = validate_and_convert_data_types(addresses, mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
        addresses = validate_and_convert_data_types(addresses, optional_columns, 'optional', 'addresses')

    if addresses is None:
        return None
    
    url = create_url("centerofgravity")
    
    headers = create_headers(api_key)
    payload = {
        "addresses": addresses.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "center of gravity")
    if response_data is None:
        return None
    else:
        assigned_addresses_df = pd.DataFrame(response_data['assignedAddresses'])
        centers_df = pd.DataFrame(response_data['centers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_addresses_df, centers_df
            
def forward_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'COGSampleDataAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:H', dtype={'postalCode': str})

    parameters = {
        "numberOfCenters": 5,
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_center_of_gravity(coordinates: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Calculate reverse center of gravity based on a list of geocodes and their weights.

    This function takes a DataFrame of coordinates with weights and a set of parameters,
    along with an API key, and performs reverse center of gravity calculation using the Log-hub service.

    Parameters:
    coordinates (pd.DataFrame): A pandas DataFrame containing geocodes with their weights.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the location.
        - longitude (float): Longitude of the location.
        - weight (float): Weight.

    parameters (Dict): A dictionary containing parameters like numberOfCenters (number) 
                       and distanceUnit (enum: "km" or "mi").

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    api_key (str): The Log-hub API key for accessing the reverse center of gravity service.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: A tuple of two pandas DataFrames. The first DataFrame 
                                       contains the assigned geocodes with their respective centers, 
                                       and the second DataFrame contains the details of the centers.
                                       Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    mandatory_columns = {'latitude': 'float', 'longitude': 'float', 'weight': 'float'}
    optional_columns = {'id': 'float', 'name': 'str'}

    # Validate and convert data types
    coordinates = validate_and_convert_data_types(coordinates, mandatory_columns, 'mandatory', 'coordinates')
    if not coordinates is None:
        coordinates = validate_and_convert_data_types(coordinates, optional_columns, 'optional', 'coordinates')

    if coordinates is None:
        return None

    url = create_url("reversecenterofgravity")
    headers = create_headers(api_key)
    payload = {
        "coordinates": coordinates.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario,payload)
    response_data = post_method(url, payload, headers, "reverse center of gravity")
    if response_data is None:
        return None
    else:
        assigned_geocodes_df = pd.DataFrame(response_data['assignedGeocodes'])
        centers_df = pd.DataFrame(response_data['centers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_geocodes_df, centers_df
            
def reverse_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'COGSampleDataReverse.xlsx')
    coordinates_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:E')

    parameters = {
        "numberOfCenters": 5,
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'coordinates': coordinates_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}



import os
import pandas as pd
import logging
from typing import Optional, Tuple
import warnings
logging.basicConfig(level=logging.INFO)
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities


def forward_nearest_warehouses(warehouses: pd.DataFrame, customers: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform forward nearest warehouses on a list of addresses.

    This function takes a DataFrame of addresses, parameters, and an API key, and performs forward
    nearest warehouses using the Log-hub service. 

    Parameters:
    warehouses (pd.DataFrame): A pandas DataFrame containing warehouses that should be considered in the calculation.
        Required columns and their types:
        - name (str): Warehouse name.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.

    customers (pd.DataFrame): A pandas DataFrame containing customers that should be considered in the calculation. 
        Required columns and their types:
        - name (str): Customer name.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.

    parametrs (dict): A dictionary containing parameters 'nearestWarehouses' (an integer number), 'distanceUnit' (enum 'km' or 'mi'), 'durationUnit' (enum 'min', 'sec' or 'h'), 'vehicleType' (enum 'car' or 'truc'), 'routePreference' (enum 'recommended', 'shortest', 'fastest'), 'maxDistance' (an integer number), and 'streetLevel' (boolean).

    api_key (str): The Log-hub API key for accessing the nearest warehouses service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Optional[Tuple[pd.DataFrame, pd.DataFrame]]: A pandas DataFrame containing information about the nearest warehouses for each 
                                                 customer, and a pandas DataFrame for unassigned customers. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    warehouses_mandatory_columns = {'name': 'str', 'country': 'str'}
    warehouses_optional_columns = {'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    customer_mandatory_columns = {'name': 'str', 'country': 'str'}
    customer_optional_columns = {'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}

    # Validate and convert data types
    warehouses = validate_and_convert_data_types(warehouses, warehouses_mandatory_columns, 'mandatory', 'warehouses')
    if not warehouses is None:
        warehouses = validate_and_convert_data_types(warehouses, warehouses_optional_columns, 'optional', 'warehouses')
    customers = validate_and_convert_data_types(customers, customer_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customer_optional_columns, 'optional', 'customers')

    if any(df is None for df in [warehouses, customers]):
        return None
    
    url = create_url("nearestwarehouses")
    
    headers = create_headers(api_key)

    payload = {
        'warehouses': warehouses.to_dict(orient='records'),
        'customers': customers.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "nearest warehouses")
    if response_data is None:
        return None
    else: 
        nearest_warehouses_df = pd.DataFrame(response_data['nearestWarehouses'])
        unassigned_df = pd.DataFrame(response_data['unassignedCustomers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return nearest_warehouses_df, unassigned_df
            
def forward_nearest_warehouses_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'nearestWarehousesAddresses.xlsx')
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:G').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:G').fillna("")
    parameters = {
        "nearestWarehouses": 3,
        "distanceUnit": "km",
        "durationUnit": "min",
        "vehicleType": "car",
        "routePreference": "recommended",
        "maxDistance": 2000,
        "streetLevel": False
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'warehouses': warehouses_df, 'customers': customers_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_nearest_warehouses(warehouses: pd.DataFrame, customers: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse nearest warehouses on a list of coordinates for warehouses and customers.

    This function takes a DataFrame of coordinates for warehouses and customers, parameters, and an API key, and performs reverse
    nearest warehouses using the Log-hub service. 

    Parameters:
    warehouses (pd.DataFrame): A pandas DataFrame containing warehouses that should be considered in the calculation.
        Required columns and their types:
        - name (str): Warehouse name.
        - latitude (number): Warehouse latitude.
        - longitude (number): Warehouse longitude.

    customers (pd.DataFrame): A pandas DataFrame containing customers that should be considered in the calculation. 
        Required columns and their types:
        - name (str): Customer name.
        - latitude (number): Customer latitude.
        - longitude (number): Customer longitude.

    parametrs (dict): A dictionary containing parameters 'nearestWarehouses' (an integer number), 'distanceUnit' (enum 'km' or 'mi'), 'durationUnit' (enum 'min', 'sec' or 'h'), 'vehicleType' (enum 'car' or 'truc'), 'routePreference' (enum 'recommended', 'shortest', 'fastest'), 'maxDistance' (an integer number), and 'streetLevel' (boolean).

    api_key (str): The Log-hub API key for accessing the nearest warehouses service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Optional[Tuple[pd.DataFrame, pd.DataFrame]]: A pandas DataFrame containing information about the nearest warehouses for each 
                                                 customer, and a pandas DataFrame for unassigned customers. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    warehouses_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float'
    }
    customer_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float'
    }

    # Validate and convert data types
    warehouses = validate_and_convert_data_types(warehouses, warehouses_columns, 'mandatory', 'warehouses')
    customers = validate_and_convert_data_types(customers, customer_columns, 'mandatory', 'customers')

    if any(df is None for df in [warehouses, customers]):
        return None
    
    url = create_url("reversenearestwarehouses")
    
    headers = create_headers(api_key)

    payload = {
        'warehouses': warehouses.to_dict(orient='records'),
        'customers': customers.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url,payload, headers, "reverse nearest warehouses")
    if response_data is None:
        return None
    else:
        nearest_warehouses_df = pd.DataFrame(response_data['nearestWarehouses'])
        unassigned_df = pd.DataFrame(response_data['unassignedCustomers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return nearest_warehouses_df, unassigned_df

def reverse_nearest_warehouses_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'nearestWarehousesReverse.xlsx')
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:D').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:D').fillna("")
    parameters = {
        "nearestWarehouses": 3,
        "distanceUnit": "km",
        "durationUnit": "min",
        "vehicleType": "car",
        "routePreference": "recommended",
        "maxDistance": 2000,
        "streetLevel": False
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'warehouses': warehouses_df, 'customers': customers_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}



import os
import pandas as pd
import logging
import warnings
logging.basicConfig(level=logging.INFO)
from typing import Optional, Tuple
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import exclude_nan_depending_on_dtype, remove_nonexisting_optional_columns
from sending_requests import post_method, create_headers, create_url, get_method, get_workspace_entities

def forward_location_planning(customers: pd.DataFrame, warehouses: pd.DataFrame, costs_adjustments: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform location planning based on customers, warehouses, and costs adjustment.

    This function takes DataFrames of customers, warehouses, costs adjustment, and an API key, and performs forward network design plus using the Log-hub service. 

    Parameters:
    customers (pd.DataFrame): Customer data with its address and its shipment structure for the optimization run.
        - name (str): Customer name.
        - country (str): Country code or country name.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - weight (number): Average weight of a shipment that goes to this customer.
        - volume (number): Average volume of a shipment that goes to this customer.
        - numberOfShipments (number): Total number of shipments (integer value) the customer receives.
    warehouses (pd.DataFrame): A pandas DataFrame containing the warehouses information for the optimization run.
        - name (str): Warehouse name.
        - country (str): Country code or country name.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - fixed (0 or 1): If the warehouse should be in every case part of the result, then mark it as fixed (1).
        - minWeight (number): The weight that should be at least assigned to this warehouse.
        - maxWeight (number): The maximum weight capacity of the warehouse.
        - penaltyCostsWeight (number): Penalty costs per weight unit. 
        - minVolume (number): The volume that should be at least assigned to this warehouse.
        - maxVolume (number): The maximum volume capacity of the warehouse.
        - penaltyCostsVolume (number): Penalty costs per volume unit. 
        - fixedCosts (number):  Fixed warehouse costs.
        - costsPerWeightUnit (number): Costs by weight unit.
        - costsPerVolumeUnit (number): Costs by volume unit.
    costs_adjustments (pd.DataFrame): A pandas DataFrame containing information that allows users to customize the transportation costs between warehouses and customers by defining adjustment factors and flat on top fees.
        - customerCountryIso2 (str): A two-letter ISO2 code representing the country of the customer.
        - warehouseCountryIso2 (str): A two-letter ISO2 code representing the country of the warehouse.
        - customerName (str): Customer name.
        - warehouseName (str): Warehouse name.
        - adjustmentFactor (number): A numerical value which will be multiplied by the transport costs.
        - flatOnTop (number): A flat fee that will be added to the transport costs.

    parametrs (dict): A dictionary containing parameters distanceUnit (enum 'km' or 'mi'), weightUnit (enum 'kg' or 't' or 'LBS'), volumeUnit (enum 'cbm' or 'l'), vehicleType (enum 'car' or 'truck'), streetLevel (boolean), minWarehouses (integer number), maxWarehouses (integer number). 

    api_key (str): The Log-hub API key for accessing the location planning service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three dataframes with the information about opened warehouses, customer assignment, and solution kpis. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    #removes missing optional columns from the list and merges optional and mandatory columns, then validates data types
    customers_mandatory_columns = {'name': 'str', 'country': 'str', 'weight': 'float', 'volume': 'float', 'numberOfShipments': 'float'}
    customers_optional_columns = {'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    sent_optional_columns = set(customers.columns) - set(customers_mandatory_columns.keys())
    customers_optional_columns = remove_nonexisting_optional_columns(customers_optional_columns, sent_optional_columns)
    customers_mandatory_columns.update(customers_optional_columns)
    customers = exclude_nan_depending_on_dtype(customers, customers_mandatory_columns, 'customers')

    warehouses_mandatory_columns = {'name': 'str', 'country': 'str', 'fixed': 'float', 'minWeight': 'float', 'maxWeight': 'float', 'minVolume': 'float', 'maxVolume': 'float', 'fixedCosts': 'float','costsPerWeightUnit': 'float', 'costsPerVolumeUnit': 'float'}
    warehouses_optional_columns = {'id': 'float', 'state' : 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'penaltyCostsWeight': 'float', 'penaltyCostsVolume': 'float'}
    sent_optional_columns = set(warehouses.columns) - set(warehouses_mandatory_columns.keys())
    warehouses_optional_columns = remove_nonexisting_optional_columns(warehouses_optional_columns, sent_optional_columns)
    warehouses_mandatory_columns.update(warehouses_optional_columns)
    warehouses = exclude_nan_depending_on_dtype(warehouses, warehouses_mandatory_columns, 'warehouses')

    costs_adjustments_optional_columns = {'id': 'float', 'customerCountryIso2': 'str',  'warehouseCountryIso2': 'str', 'customerName': 'str', 'warehouseName': 'str', 'adjustmentFactor': 'float', 'flatOnTop': 'float'}
    sent_optional_columns = set(costs_adjustments.columns)
    costs_adjustments_optional_columns = remove_nonexisting_optional_columns(costs_adjustments_optional_columns, sent_optional_columns)
    costs_adjustments = exclude_nan_depending_on_dtype(costs_adjustments, costs_adjustments_optional_columns, 'costs_adjustments')

    if any(df is None for df in [warehouses, customers, costs_adjustments]):
        return None
    
    url = create_url("locationplanninglongrun")
    
    headers = create_headers(api_key)

    payload = {
        'customers': customers,
        'warehouses': warehouses,
        'costsAdjustmentTransportationCostsStandard': costs_adjustments,
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "location planning")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "location planning")
        if get_method_result is None:
            return None
        else:
            open_warehouses = pd.DataFrame(get_method_result['openWarehouses'])
            customer_assignment = pd.DataFrame(get_method_result['customerAssignment'])
            solution_kpis = pd.DataFrame(get_method_result['solutionKpis'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return open_warehouses, customer_assignment,solution_kpis

def forward_location_planning_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'LocationPlanningAddresses.xlsx')
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:Q').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:J').fillna("")
    costs_adjustments_df = pd.read_excel(data_path, sheet_name='costsAdjustment', usecols='A:G').fillna("")
    parameters = {
        "distanceUnit": "km",
        "vehicleType": "car",
        "streetLevel": False,
        "weightUnit": "kg",
        "volumeUnit": "cbm",
        "minWarehouses": 5,
        "maxWarehouses": 6

    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'warehouses': warehouses_df, 'customers': customers_df, 'costsAdjustments': costs_adjustments_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_location_planning(customers: pd.DataFrame, warehouses: pd.DataFrame, costs_adjustments: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse location planning based on customers, warehouses, and costs adjustment.

    This function takes DataFrames of customers, warehouses, costs adjustment, and an API key, and performs reverse network design plus using the Log-hub service. 

    Parameters:
    customers (pd.DataFrame): Customer data with its address and its shipment structure for the optimization run.
        - name (str): Customer name.
        - latitude (number): Customer latitude.
        - longitude (number): Customer longitude.
        - weight (number): Average weight of a shipment that goes to this customer.
        - volume (number): Average volume of a shipment that goes to this customer.
        - numberOfShipments (number): Total number of shipments (integer value) the customer receives.
    warehouses (pd.DataFrame): A pandas DataFrame containing the warehouses information for the optimization run.
        - name (str): Warehouse name.
        - latitude (number): Warehouse latitude.
        - longitude (number): Warehouse longitude.
        - fixed (0 or 1): If the warehouse should be in every case part of the result, then mark it as fixed (1).
        - minWeight (number): The weight that should be at least assigned to this warehouse.
        - maxWeight (number): The maximum weight capacity of the warehouse.
        - penaltyCostsWeight (number): Penalty costs per weight unit. 
        - minVolume (number): The volume that should be at least assigned to this warehouse.
        - maxVolume (number): The maximum volume capacity of the warehouse.
        - penaltyCostsVolume (number): Penalty costs per volume unit. 
        - fixedCosts (number):  Fixed warehouse costs.
        - costsPerWeightUnit (number): Costs by weight unit.
        - costsPerVolumeUnit (number): Costs by volume unit.
    costs_adjustments (pd.DataFrame): A pandas DataFrame containing information that allows users to customize the transportation costs between warehouses and customers by defining adjustment factors and flat on top fees.
        - customerCountryIso2 (str): A two-letter ISO2 code representing the country of the customer.
        - warehouseCountryIso2 (str): A two-letter ISO2 code representing the country of the warehouse.
        - customerName (str): Customer name.
        - warehouseName (str): Warehouse name.
        - adjustmentFactor (number): A numerical value which will be multiplied by the transport costs.
        - flatOnTop (number): A flat fee that will be added to the transport costs.

    parametrs (dict): A dictionary containing parameters distanceUnit (enum 'km' or 'mi'), weightUnit (enum 'kg' or 't' or 'LBS'), volumeUnit (enum 'cbm' or 'l'), vehicleType (enum 'car' or 'truck'), streetLevel (boolean), minWarehouses (integer number), maxWarehouses (integer number). 

    api_key (str): The Log-hub API key for accessing the location planning service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three dataframes with the information about opened warehouses, customer assignment, and solution kpis. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    #removes missing optional columns from the list and merges optional and mandatory columns, then validates data types
    customers_mandatory_columns = {'name': 'str', 'latitude': 'float', 'longitude': 'float', 'weight': 'float', 'volume': 'float', 'numberOfShipments': 'float'}
    customers_optional_columns = {'id': 'float'}
    sent_optional_columns = set(customers.columns) - set(customers_mandatory_columns.keys())
    customers_optional_columns = remove_nonexisting_optional_columns(customers_optional_columns, sent_optional_columns)
    customers_mandatory_columns.update(customers_optional_columns)
    customers = exclude_nan_depending_on_dtype(customers, customers_mandatory_columns, 'customers')

    warehouses_mandatory_columns = {'name': 'str', 'latitude': 'float', 'longitude': 'float', 'fixed': 'float', 'minWeight': 'float', 'maxWeight': 'float', 'minVolume': 'float', 'maxVolume': 'float', 'fixedCosts': 'float','costsPerWeightUnit': 'float', 'costsPerVolumeUnit': 'float'}
    warehouses_optional_columns = {'id': 'float', 'penaltyCostsWeight': 'float', 'penaltyCostsVolume': 'float'}
    sent_optional_columns = set(warehouses.columns) - set(warehouses_mandatory_columns.keys())
    warehouses_optional_columns = remove_nonexisting_optional_columns(warehouses_optional_columns, sent_optional_columns)
    warehouses_mandatory_columns.update(warehouses_optional_columns)
    warehouses = exclude_nan_depending_on_dtype(warehouses, warehouses_mandatory_columns, 'warehouses')

    costs_adjustments_optional_columns = {'id': 'float', 'customerCountryIso2': 'str',  'warehouseCountryIso2': 'str', 'customerName': 'str', 'warehouseName': 'str', 'adjustmentFactor': 'float', 'flatOnTop': 'float'}
    sent_optional_columns = set(costs_adjustments.columns)
    costs_adjustments_optional_columns = remove_nonexisting_optional_columns(costs_adjustments_optional_columns, sent_optional_columns)
    costs_adjustments = exclude_nan_depending_on_dtype(costs_adjustments, costs_adjustments_optional_columns, 'costs_adjustments')

    if any(df is None for df in [warehouses, customers, costs_adjustments]):
        return None
    
    url = create_url("reverselocationplanninglongrun")
    
    headers = create_headers(api_key)

    payload = {
        'customers': customers,
        'warehouses': warehouses,
        'costsAdjustmentTransportationCostsStandard': costs_adjustments,
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "reverse location planning")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "reverse location planning")
        if get_method_result is None:
            return None
        else:
            open_warehouses = pd.DataFrame(get_method_result['openWarehouses'])
            customer_assignment = pd.DataFrame(get_method_result['customerAssignment'])
            solution_kpis = pd.DataFrame(get_method_result['solutionKpis'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return open_warehouses, customer_assignment, solution_kpis

def reverse_location_planning_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'LocationPlanningReverse.xlsx')
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:N').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:G').fillna("")
    costs_adjustments_df = pd.read_excel(data_path, sheet_name='costsAdjustment', usecols='A:G').fillna("")
    parameters = {
        "distanceUnit": "km",
        "vehicleType": "car",
        "streetLevel": False,
        "weightUnit": "kg",
        "volumeUnit": "cbm",
        "minWarehouses": 5,
        "maxWarehouses": 6

    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'warehouses': warehouses_df, 'customers': customers_df, 'costsAdjustments': costs_adjustments_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_supply_chain_map_relations(addresses: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of relations based on the given addresses.

    This function takes a DataFrame of addresses pairs with their layers and quantity, along with an API key, 
    and creates a map of relations using the Log-hub Supply Chain Map service. 

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing addresses pairs with their layers and quantity.
        Each row should contain:
        - id (number): Identifier.
        - senderName (str): Name for the start location of the relation.
        - senderCountry (str): Country code or country name for the start location of the relation. Must have at least 
                               1 character.
        - senderState (str): State code for the start location of the relation.
        - senderPostalCode (str): Postal code for the start location of the relation.
        - senderSity (str): City name for the start location of the relation.
        - senderStreet (str): Street name with house number for the start location of the relation.
        - senderLocationLayer (str): Layer name for the start location.
        - recipientName (str): Name for the end location of the relation.
        - recipientCountry (str): Country code or country name for the end location of the relation. Must have at least 
                               1 character.
        - recipientState (str): State code for the end location of the relation.
        - recipientPostalCode (str): Postal code for the end location of the relation.
        - recipientSity (str): City name for the end location of the relation.
        - recipientStreet (str): Street name with house number for the end location of the relation.
        - recipientLocationLayer (str): Layer name for the end location.
        - relationLayer (str): Layer name for the relation.
        - quantity (number): Numerical value affecting the thickness of the relation on the map.

    parameters (dict): A dictionary containing parameter 'showLocations' (boolean).

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 'workspaceId' (str) and 'scenarioName' (str).
                    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containg the addresses pairs and their parsed latitude and longitude. Returns None if 
                  the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    mandatory_columns = {'senderCountry': 'str',  'recipientCountry': 'str'}
    optional_columns = {'id': 'float', 'senderName': 'str', 'senderState': 'str', 'senderPostalCode': 'str', 'senderCity': 'str', 'senderStreet': 'str', 'senderLocationLayer': 'str', 'recipientName': 'str', 'recipientState': 'str', 'recipientPostalCode': 'str', 'recipientCity': 'str', 'recipientStreet': 'str', 'recipientLocationLayer': 'str', 'relationLayer': 'str', 'quantity': 'string'}

    # Validate and convert data types
    addresses = validate_and_convert_data_types(addresses, mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
        addresses = validate_and_convert_data_types(addresses, optional_columns, 'optional', 'addresses')
    if addresses is None:
        return None

    url = create_url("supplychainmaprelations")
    headers = create_headers(api_key)
    payload = {
        "distanceCalculationData": addresses.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload, "supplychainmaprelations")
    
    response_data = post_method(url, payload, headers,"supply chain map relations")
    if response_data is None:
        return None
    else:
        distance_calculation_result_df = pd.DataFrame(response_data['distanceCalculationResult'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return distance_calculation_result_df

def forward_supply_chain_map_relations_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapRelationsAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:Q', dtype={'postalCode': str})

    parameters = {
        'showLocations': True
    }

    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_supply_chain_map_relations(coordinates: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of relations based on the given coordinates pairs.

    This function takes a DataFrame of coordinates pairs with their layers and quantity, along with an API key, 
    and creates a map of relations using the Log-hub Supply Chain Map service. 

    Parameters:
    coordinates (pd.DataFrame): A pandas DataFrame containing coordinates pairs with their layers and quantity.
        Each row should contain:
        - id (number): Identifier.
        - senderName (str): Name for the start location of the relation.
        - senderLatitude (number): Latitude for the start location of the relation.
        - senderLongitude (number): Longitude for the start location of the relation.
        - senderLocationLayer (str): Layer name for the start location.
        - recipientName (str): Name for the end location of the relation.
        - recipientLatitude (number): Latitude for the end location of the relation.
        - recipientLongitude (number): Longitude for the end location of the relation.
        - recipientLocationLayer (str): Layer name for the end location.
        - relationLayer (str): Layer name for the relation.
        - quantity (number): Numerical value affecting the thickness of the relation on the map.

    parameters (dict): A dictionary containing parameter 'showLocations' (boolean).

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 'workspaceId' (str) and 'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containg the coordinates pairs. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    mandatory_columns = {
            'senderLatitude':'float', 'senderLongitude': 'float', 'recipientLatitude':'float', 'recipientLongitude': 'float'
        }
    optional_columns = {'id': 'float', 'senderName': 'str', 'senderLocationLayer': 'str', 'recipientName': 'str', 'recipientLocationLayer': 'str', 'relationLayer': 'str', 'quantity': 'str'}
    # Validate and convert data types
    coordinates = validate_and_convert_data_types(coordinates, mandatory_columns, 'mandatory', 'coordinates')
    if not coordinates is None:
        coordinates = validate_and_convert_data_types(coordinates, optional_columns, 'optional', 'coordinates')
    if coordinates is None:
        return None

    url = create_url("reversesupplychainmaprelations")
    
    headers = create_headers(api_key)
    payload = {
        "distanceCalculationData": coordinates.to_dict(orient='records'),
        'parameters': parameters
    }
    
    payload = save_scenario_check(save_scenario, payload, "reversesupplychainmaprelations")
    
    response_data = post_method(url, payload, headers, "reverse supply chain map relations")
    if response_data is None:
        return None
    else:
        distance_calculation_result_df = pd.DataFrame(response_data['distanceCalculationData'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return distance_calculation_result_df

def reverse_supply_chain_map_relations_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapRelationsReverse.xlsx')
    coordinates_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:K')

    parameters = {
        'showLocations': True
    }

    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'coordinates': coordinates_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os

import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import convert_timestamps, validate_and_convert_data_types, convert_to_float, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_transport_optimization_plus(vehicles: pd.DataFrame, jobs: pd.DataFrame, timeWindowProfiles: pd.DataFrame, breaks: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform transport optimization based on vehicles, jobs, time window profiles, and breaks.

    Parameters:
    vehicles (pd.DataFrame): DataFrame containing vehicle information.
        Columns:
        - vehicleTypeId (str): Vehicle Type Id.
        - availableVehicles (int): Available Vehicles.
        - [startId, endId] (str): Start and End Ids.
        - [startCountry, endCountry] (str): Start and End Country.
        - [startState, endState] (str): Start and End State.
        - [startPostalCode, endPostalCode] (str): Start and End Postal Codes.
        - [startCity, endCity] (str): Start and End City.
        - [startStreet, endStreet] (str): Start and End Street.
        - maxWeight (float): Maximum Weight.
        - maxVolume (float): Maximum Volume.
        - maxPallets (int): Maximum Pallets.
        - maxStops (int): Maximum Stops.
        - timeWindowStart, timeWindowEnd (str): Time Window Start and End (ISO 8601 format).
        - profile (str): Profile.
        - speedFactor (float): Speed Factor.
        - fixed, perHour, perKilometer, costPerStop (float): Fixed and Costs Per Hour, Kilometer and Stop.
        - minimumTravelTime, maxTravelTime (float): Minimum and Maximum Travel Time.
        - max_distance (float): Maximum number of kilometers that vehicle can drive per route.
        - maximumDistanceBetweenStops (float): Maximum number of kilometers that vehicle can drive between each stop.
        - breakId (str): Break Id.

    jobs (pd.DataFrame): DataFrame containing job information.
        Columns:
        - shipmentId (str): Shipment Id.
        - [fromName, toName] (str): From and To Names.
        - [fromCountry, toCountry] (str): From and To Countries.
        - [fromState, toState] (str): From and To States.
        - [fromPostalCode, toPostalCode] (str): From and To Postal Codes.
        - [fromCity, toCity] (str): From and To Cities.
        - [fromStreet, toStreet] (str): From and To Streets.
        - [senderStopDuration, recipientStopDuration] (float): Sender and Recipient Stop Durations.
        - [senderTimeWindowProfile, recipientTimeWindowProfile] (str): Sender and Recipient Time Window Profiles.
        - weight, volume (float): Weight and Volume.
        - pallets (int): Pallets.
        - vehicleTypeId (str): Vehicle Type Id.
        - external_costs (number): External costs for order.

    timeWindowProfiles (pd.DataFrame): DataFrame containing time window profile information.
        Columns:
        - timeWindowProfileId (str): Time Window Profile Id.
        - timeWindowProfileStart, timeWindowProfileEnd (str): Time Window Profile Start and End (ISO 8601 format).

    breaks (pd.DataFrame): DataFrame containing break information.
        Columns:
        - breakId (str): Break Id.
        - earliestBreakStart, latestBreakStart (str): Earliest and Latest Break Start (ISO 8601 format).
        - earliestRelativeBreakStart, latestRelativeBreakStart (str): Earliest and Latest Relative Break Start (ISO 8601 format).
        - breakDuration (float): Break Duration.

    parameters (Dict): Dictionary containing parameters like durationUnit and distanceUnit.

    api_key (str): Log-hub API key for accessing the transport optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three pandas DataFrames containing route overview, route details, 
                                                     and external orders. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    # Define expected columns and data types for each DataFrame
    vehicle_mandatory_columns = {
        'vehicleTypeId': 'str', 'availableVehicles': 'int'
    }
    vehicle_optional_columns = {
        'startId': 'str', 'startCountry': 'str', 'startState': 'str', 'startPostalCode': 'str', 'startCity': 'str', 'startStreet': 'str', 'endId': 'str', 
        'endCountry': 'str', 'endState': 'str', 'endPostalCode': 'str', 'endCity': 'str', 'endStreet': 'str', 
        'timeWindowStart': 'str', 'timeWindowEnd': 'str', 'profile': 'str', 'breakId': 'str'
    }
    vehicle_optional_floats = ['maxWeight', 'maxVolume', 'maxPallets', 'maxStops', 'speedFactor', 'fixed', 'perHour', 'perKilometer', 'costPerStop', 'minimumTravelTime', 'maxTravelTime', 'max_distance', 'maximumDistanceBetweenStops']
    job_mandatory_columns = {
        'shipmentId': 'str', 'fromName': 'str', 'fromCountry': 'str', 'toName': 'str', 'toCountry': 'str', 'vehicleTypeId': 'str'
    }
    job_optional_columns = {
        'fromState': 'str', 'fromPostalCode': 'str', 'fromCity': 'str', 'fromStreet': 'str', 'senderTimeWindowProfile': 'str', 'toState': 'str', 'toPostalCode': 'str', 'toCity': 'str', 
        'toStreet': 'str', 'recipientTimeWindowProfile': 'str',
    }
    job_optional_floats = ['senderStopDuration', 'recipientStopDuration', 'weight', 'volume', 'pallets', 'external_costs']
    timeWindowProfile_optional_columns = {
        'timeWindowProfileId': 'str', 'timeWindowProfileStart': 'str', 'timeWindowProfileEnd': 'str'
    }
    break_optional_columns = {
        'breakId': 'str', 'earliestBreakStart': 'str', 'latestBreakStart': 'str', 'earliestRelativeBreakStart': 'str', 'latestRelativeBreakStart': 'str'
    }
    break_optional_floats = ['breakDuration']

    # Perform validation and conversion for each DataFrame
    vehicles = validate_and_convert_data_types(vehicles, vehicle_mandatory_columns, 'mandatory', 'vehicles')
    if not vehicles is None:
        vehicles = validate_and_convert_data_types(vehicles, vehicle_optional_columns, 'optional', 'vehicles')
        if not vehicles is None:
            vehicles = convert_timestamps(vehicles)
            vehicles = convert_to_float(vehicles, vehicle_optional_floats, 'optional')
            vehicles = convert_df_to_dict_excluding_nan(vehicles, vehicle_optional_floats)
    jobs = validate_and_convert_data_types(jobs, job_mandatory_columns, 'mandatory', 'jobs')
    if not jobs is None:
        jobs = validate_and_convert_data_types(jobs, job_optional_columns, 'optional', 'jobs')
        if not jobs is None:
            jobs = convert_timestamps(jobs)
            jobs = convert_to_float(jobs, job_optional_floats, 'optional')
            jobs = convert_df_to_dict_excluding_nan(jobs, job_optional_floats)
    timeWindowProfiles = validate_and_convert_data_types(timeWindowProfiles, timeWindowProfile_optional_columns, 'optional', 'time window profiles')
    if not timeWindowProfiles is None:
        timeWindowProfiles = convert_timestamps(timeWindowProfiles)
    breaks = validate_and_convert_data_types(breaks, break_optional_columns, 'optional', 'breaks')
    if not breaks is None:
        breaks = convert_timestamps(breaks)
        breaks = convert_to_float(breaks, break_optional_floats, 'optional')
        breaks = convert_df_to_dict_excluding_nan(breaks, break_optional_floats)

    # Exit if any DataFrame validation failed
    if any(df is None for df in [vehicles, jobs, timeWindowProfiles, breaks]):
        return None

    url = create_url("transportoptimizationplus")
    
    headers = create_headers(api_key)

    payload = {
        "vehicles": vehicles,
        "jobs": jobs,
        "timeWindowProfiles": timeWindowProfiles.to_dict(orient='records'),
        "breaks": breaks,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    
    response_data = post_method(url, payload, headers, "transport optimization plus")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['externalOrders'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df
           
def forward_transport_optimization_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'transportPlusAddresses.xlsx')
    vehicles_df = pd.read_excel(data_path, sheet_name='vehicles', usecols='A:AF', dtype={'startState': str, 'startPostalCode': str, 'endState': str, 'endPostalCode': str, 'maxTravelTime': int}).fillna("")
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:W').fillna("")
    time_window_profiles_df = pd.read_excel(data_path, sheet_name='timeWindowProfile', usecols='A:D').fillna("")
    breaks_df = pd.read_excel(data_path, sheet_name='breaks', usecols='A:G').fillna("")

    parameters = {
        "durationUnit": "min",
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'vehicles': vehicles_df, 'shipments': shipments_df, 'timeWindowProfiles': time_window_profiles_df,
             'breaks': breaks_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_transport_optimization_plus(vehicles: pd.DataFrame, jobs: pd.DataFrame, timeWindowProfiles: pd.DataFrame, breaks: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse transport optimization based on vehicles, jobs, time window profiles, and breaks.

    Parameters:
    vehicles (pd.DataFrame): DataFrame containing vehicle information with latitude and longitude.
        Columns:
        - vehicleTypeId (str): Vehicle Type Id.
        - availableVehicles (int): Number of available vehicles.
        - [startId, endId] (str): Start and End Ids.
        - [startLatitude, endLatitude] (float): Start and End Latitude.
        - [startLongitude, endLongitude] (float): Start and End Longitude.
        - maxWeight (float): Maximum Weight.
        - maxVolume (float): Maximum Volume.
        - maxPallets (int): Maximum Pallets.
        - maxStops (int): Maximum Stops.
        - timeWindowStart, timeWindowEnd (str): Time Window Start and End (ISO 8601 format).
        - profile (str): Profile.
        - speedFactor (float): Speed Factor.
        - fixed, perHour, perKilometer, costPerStop (float): Fixed and Costs Per Hour, Kilometer and Stop.
        - minimumTravelTime, maxTravelTime (float): Minimum and Maximum Travel Time.
        - max_distance (float): Maximum number of kilometers that vehicle can drive per route.
        - maximumDistanceBetweenStops (float): Maximum number of kilometers that vehicle can drive between each stop.
        - breakId (str): Break Id.

    jobs (pd.DataFrame): DataFrame containing job information with latitude and longitude.
        Columns:
        - shipmentId (str): Shipment Id.
        - [fromName, toName] (str): From and To Names.
        - [fromLatitude, toLatitude] (float): From and To Latitude.
        - [fromLongitude, toLongitude] (float): From and To Longitude.
        - [senderStopDuration, recipientStopDuration] (float): Sender and Recipient Stop Durations.
        - [senderTimeWindowProfile, recipientTimeWindowProfile] (str): Sender and Recipient Time Window Profiles.
        - weight, volume (float): Weight and Volume.
        - pallets (int): Pallets.
        - vehicleTypeId (str): Vehicle Type Id.
        - external_costs (number): External costs for order.

    timeWindowProfiles (pd.DataFrame): DataFrame containing time window profile information.
        Columns:
        - timeWindowProfileId (str): Time Window Profile Id.
        - timeWindowProfileStart, timeWindowProfileEnd (str): Time Window Profile Start and End (ISO 8601 format).

    breaks (pd.DataFrame): DataFrame containing break information.
        Columns:
        - breakId (str): Break Id.
        - earliestBreakStart, latestBreakStart (str): Earliest and Latest Break Start (ISO 8601 format).
        - earliestRelativeBreakStart (str): Minimum driving time before the break can start (ISO 8601 format).
        - latestRelativeBreakTime (str): Maximum driving time until the break must start (ISO 8601 format).
        - breakDuration (float): Break Duration.

    parameters (Dict): Dictionary containing parameters like durationUnit and distanceUnit.

    api_key (str): Log-hub API key for accessing the reverse transport optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three pandas DataFrames containing route overview, route details, 
                                                     and external orders. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    # Define expected columns and data types for each DataFrame
    vehicle_mandatory_columns = {
        'vehicleTypeId': 'str', 'availableVehicles': 'int'
    }
    vehicle_optional_columns = {
        'startId': 'str', 'endId': 'str', 'timeWindowStart': 'str', 'timeWindowEnd': 'str', 'profile': 'str', 'breakId': 'str'
    }
    vehicle_optional_floats = ['startLatitude', 'startLongitude', 'endLatitude', 'endLongitude', 'maxWeight', 'maxVolume', 'maxPallets', 'maxStops', 'speedFactor', 'fixed', 'perHour',  'perKilometer', 'costPerStop', 'minimumTravelTime', 'maxTravelTime', 'max_distance', 'maximumDistanceBetweenStops']
    job_mandatory_columns = {
        'shipmentId': 'str', 'fromName': 'str', 'fromLatitude': 'float', 'fromLongitude': 'float', 'toName': 'str', 'toLatitude': 'float', 'toLongitude': 'float'
    }
    job_optional_columns = {
        'senderTimeWindowProfile': 'str', 'recipientTimeWindowProfile': 'str', 'vehicleTypeId': 'str',
    }
    job_optional_floats = ['senderStopDuration', 'recipientStopDuration', 'weight', 'volume', 'pallets', 'external_costs']
    timeWindowProfile_optional_columns = {
        'timeWindowProfileId': 'str', 'timeWindowProfileStart': 'str', 'timeWindowProfileEnd': 'str'
    }
    break_optional_columns = {
        'breakId': 'str', 'earliestBreakStart': 'str', 'latestBreakStart': 'str', 'earliestRelativeBreakStart': 'str', 'latestRelativeBreakStart': 'str'
    }
    break_optional_floats = ['breakDuration']

    # Perform validation and conversion for each DataFrame
    vehicles = validate_and_convert_data_types(vehicles, vehicle_mandatory_columns, 'mandatory', 'vehicles')
    if not vehicles is None:
        vehicles = validate_and_convert_data_types(vehicles, vehicle_optional_columns, 'optional', 'vehicles')
        if not vehicles is None:
            vehicles = convert_timestamps(vehicles)
            vehicles = convert_to_float(vehicles, vehicle_optional_floats, 'optional')
            vehicles = convert_df_to_dict_excluding_nan(vehicles, vehicle_optional_floats)
    jobs = validate_and_convert_data_types(jobs, job_mandatory_columns, 'mandatory', 'jobs')
    if not jobs is None:
        jobs = validate_and_convert_data_types(jobs, job_optional_columns, 'optional', 'jobs')
        if not jobs is None:
            jobs = convert_timestamps(jobs)
            jobs = convert_to_float(jobs, job_optional_floats, 'optional')
            jobs = convert_df_to_dict_excluding_nan(jobs, job_optional_floats)
    timeWindowProfiles = validate_and_convert_data_types(timeWindowProfiles, timeWindowProfile_optional_columns, 'optional', 'time window profiles')
    if not timeWindowProfiles is None:
        timeWindowProfiles = convert_timestamps(timeWindowProfiles)
    breaks = validate_and_convert_data_types(breaks, break_optional_columns, 'optional', 'breaks')
    if not breaks is None:
        breaks = convert_timestamps(breaks)
        breaks = convert_to_float(breaks, break_optional_floats, 'optional')
        breaks = convert_df_to_dict_excluding_nan(breaks, break_optional_floats)

    # Exit if any DataFrame validation failed
    if any(df is None for df in [vehicles, jobs, timeWindowProfiles, breaks]):
        return None

    url = create_url("reversetransportoptimizationplus")
    
    headers = create_headers(api_key)
    payload = {
        "vehicles": vehicles,
        "jobs": jobs,
        "timeWindowProfiles": timeWindowProfiles.to_dict(orient='records'),
        "breaks": breaks,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "reverse transport optimization plus")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['externalOrders'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df

def reverse_transport_optimization_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'transportPlusReverse.xlsx')
    vehicles_df = pd.read_excel(data_path, sheet_name='vehicles', usecols='A:Z', dtype={'maxTravelTime': int}).fillna("")
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:Q').fillna("")
    time_window_profiles_df = pd.read_excel(data_path, sheet_name='timeWindowProfiles', usecols='A:D').fillna("")
    breaks_df = pd.read_excel(data_path, sheet_name='breaks', usecols='A:G').fillna("")

    parameters = {
        "durationUnit": "min",
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'vehicles': vehicles_df, 'shipments': shipments_df, 'timeWindowProfiles': time_window_profiles_df,
             'breaks': breaks_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import logging
from typing import Optional
import warnings
logging.basicConfig(level=logging.INFO)
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_geocoding(addresses: pd.DataFrame, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Perform forward geocoding on a list of addresses.

    This function takes a DataFrame of addresses and an API key, and performs forward
    geocoding using the Log-hub geocoding service. 

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing the addresses to be geocoded.
        Required columns and their types:
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - searchString (str): Key location details. Information such as country, state, city, street, and zip code, although the zip code cannot be the first entry.

    api_key (str): The Log-hub API key for accessing the geocoding service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containing the original address information along 
                  with the geocoded results. Includes latitude and longitude for each address.
                  Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    mandatory_columns = {'country': 'str'}
    optional_columns = {'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'searchString': 'str'}

    # Validate and convert data types
    addresses = validate_and_convert_data_types(addresses, mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
        addresses = validate_and_convert_data_types(addresses, optional_columns, 'optional', 'addresses')

    if addresses is None:
        return None
    
    url = create_url("geocoding")
    
    headers = create_headers(api_key)

    payload = {
        'addresses': addresses.to_dict(orient='records'),
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "geocoding")
    if response_data is None:
        return None
    else:
        geocoded_data_df = pd.DataFrame(response_data['geocodes'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return geocoded_data_df

def forward_geocoding_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'GeocodingSampleDataAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:F').fillna("")
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'saveScenarioParameters': save_scenario}

def reverse_geocoding(geocodes: pd.DataFrame, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Perform reverse geocoding on a list of latitude and longitude coordinates.

    This function takes a DataFrame of geocodes (latitude and longitude) and an API key, 
    and performs reverse geocoding using the Log-hub reverse geocoding service. 

    Parameters:
    geocodes (pd.DataFrame): A pandas DataFrame containing the geocodes to be reverse geocoded.
        Required columns and their types:
        - latitude (float): Latitude of the location.
        - longitude (float): Longitude of the location.

    api_key (str): The Log-hub API key for accessing the reverse geocoding service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containing the original geocode information along 
                  with the reverse geocoded address results. Includes fields like country, 
                  state, city, and street. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    geocodes_columns = {'latitude': 'float', 'longitude': 'float'}

    # Validate and convert data types
    geocodes = validate_and_convert_data_types(geocodes, geocodes_columns, 'mandatory', 'geocodes')
    if geocodes is None:
        return None
    
    url = create_url("reversegeocoding")
    headers = create_headers(api_key)
    payload = {
        "geocodes": geocodes.to_dict(orient='records')
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "reverse geocoding")
    if response_data is None:
        return None
    else:
        addresses_df = pd.DataFrame(response_data['addresses'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return addresses_df

def reverse_geocoding_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'GeocodingSampleDataReverse.xlsx')
    geocodes_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:B')
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'geocodes': geocodes_df, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types, convert_dates
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def demand_forecasting(past_demand_data: pd.DataFrame, future_impact_factors: pd.DataFrame, sku_parameters: pd.DataFrame, api_key: str) -> Optional[pd.DataFrame]:
    """
    Perform demand forecasting based on past demand data, future impact factors and sku parameters.
    Parameters:
    past_demand_data (pd.DataFrame): DataFrame containing the historical demand information for each SKU.
        Columns:
        - sku (str): Unique identifier for each SKU.
        - date (str): A timestamp of each demand observation.
        - demand (number): Past demand value for the SKU on the specified date.
        - [futureImpactFactor1, futureImpactFactor2, futureImpactFactor3] (str): Custom columns representing external factors that may affect demand. There can be more than three columns, depending on the number of known external factors. Supported data types are Binary Indicators (0 or 1 values indicating the absence or presence of an event e.g., promotion) and Continuous Variables (numeric values representing the intensity of an impact factor e.g., promotional spend).

    future_impact_factors (pd.DataFrame): DataFrame providing future values for the impact factors (external regressors) defined in the past_demand_data.
        Columns:
        - sku (str): Unique identifier for each SKU.
        - date (str): A timestamp of each demand observation.
        - [futureImpactFactor1, futureImpactFactor2, futureImpactFactor3] (str): Custom columns representing external factors that may affect demand. There can be more than three columns, depending on the number of known external factors. Supported data types are Binary Indicators (0 or 1 values indicating the absence or presence of an event e.g., promotion) and Continuous Variables (numeric values representing the intensity of an impact factor e.g., promotional spend).

    sku_parameters (pd.DataFrame): DataFrame containing SKU-specific parameters that customize the forecasting model for each product.
        Columns:
        - sku (str): Unique identifier for each SKU.
        - forecastPeriods (number): Defining how many periods into the future should be forecasted. Each period is one day.
        - demandFrequency (str): Showing if the demand is daily, weekly, monthly or yearly.
        - lowerTrendLimit (str): A minimum expected demand level.
        - upperTrendLimit (str): A maximum expected demand level.
        - seasonality (str): Defining a seasonality pattern. Allowed entries are weekly, monthly or yearly.
        - seasonalityType (str): Specifying how seasonal components interact with the trend. Allowed values are 'additive' and 'multiplicative'
        - confidenceInterval (str): determining the level of certanty that the demand will fall within a specific range.
        - [futureImpactFactor1, futureImpactFactor2, futureImpactFactor3] (str): Custom columns representing external factors that may affect demand. There can be more than three columns, depending on the number of known external factors. Supported data types are Binary Indicators (0 or 1 values indicating the absence or presence of an event e.g., promotion) and Continuous Variables (numeric values representing the intensity of an impact factor e.g., promotional spend).

    api_key (str): Log-hub API key for accessing the demand forecasting service.

    Returns:
    pd.DataFrame: A DataFrame containing demands prediction. Returns None if the process fails.
    """
    # Define expected columns and data types for each DataFrame
    past_demand_data_mandatory_columns = {'sku': 'str', 'date': 'str', 'demand': 'float'}
    past_demand_data_optional_columns = {'futureImpactFactor1': 'str', 'futureImpactFactor2': 'str', 'futureImpactFactor3': 'str'}
    future_impact_factors_optional_columns = {'sku': 'str', 'date': 'str', 'futureImpactFactor1': 'str', 'futureImpactFactor2': 'str', 'futureImpactFactor3': 'str'}
    sku_parameters_mandatory_columns = {'sku': 'str', 'forecastPeriods': 'float'}
    sku_parameters_optional_columns = {'demandFrequency': 'str', 'lowerTrendLimit':'str', 'upperTrendLimit': 'str', 'seasonality': 'str', 'seasonalityType': 'str', 'confidenceInterval': 'str', 'futureImpactFactor1': 'str', 'futureImpactFactor2': 'str', 'futureImpactFactor3': 'str'}

    # Perform validation and conversion for each DataFrame
    past_demand_data = validate_and_convert_data_types(past_demand_data, past_demand_data_mandatory_columns, 'mandatory', 'past demand data')
    if not past_demand_data is None:
        past_demand_data = validate_and_convert_data_types(past_demand_data, past_demand_data_optional_columns, 'optional', 'past demand data')
    future_impact_factors = validate_and_convert_data_types(future_impact_factors, future_impact_factors_optional_columns, 'optional', 'future impact factors')
    sku_parameters = validate_and_convert_data_types(sku_parameters, sku_parameters_mandatory_columns, 'mandatory', 'future impact factors')
    if not sku_parameters is None:
        sku_parameters = validate_and_convert_data_types(sku_parameters, sku_parameters_optional_columns, 'optional', 'sku parameters')
    if not any(df is None for df in [past_demand_data, future_impact_factors]):
        past_demand_data = convert_dates(past_demand_data, ['date'])
        if 'date' in future_impact_factors.columns:
            future_impact_factors = convert_dates(future_impact_factors, ['date'])

    # Exit if any DataFrame validation failed
    if any(df is None for df in [past_demand_data, future_impact_factors, sku_parameters]):
        return None

    url = create_url("demandforecasting")
    
    headers = create_headers(api_key)

    payload = {
        "pastDemandData": past_demand_data.to_dict(orient='records'),
        "futureImpactFactors": future_impact_factors.to_dict(orient='records'),
        "skuParameters": sku_parameters.to_dict(orient='records'),
    }
    
    response_data = post_method(url, payload, headers, "demand forecasting")
    if response_data is None:
        return None
    else:
        prediction_df = pd.DataFrame(response_data['prediction'])
        return prediction_df
            

def demand_forecasting_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'demandForecasting.xlsx')
    past_demand_data_df = pd.read_excel(data_path, sheet_name='pastDemandData', usecols='A:G').fillna("")
    future_impact_factors_df = pd.read_excel(data_path, sheet_name='futureImpactFactors', usecols='A:F').fillna("")
    sku_parameters_df = pd.read_excel(data_path, sheet_name='skuParameters', usecols='A:L').fillna("")

    return {'pastDemandData': past_demand_data_df, 'futureImpactFactors': future_impact_factors_df, 'skuParameters': sku_parameters_df}